/*
 * linux/kernel/chr_drv/scanner.c
 *
 *	Scanner driver for GS4500: version 1.0
 *
 *	Depending on how much scanner support we want, the
 *	following should be done:
 *
 *	    Split this file into device independent code (scanner.c)
 *	    and device dependent stuff (gs4500.c, artec.c, etc).
 *
 *	    Get an official major # allocated by RICK@EE.UWM.EDU.
 *
 *	    Create include/scanner.h with major/minors and
 *	    struct scan_info.
 *
 *	    Add configuration to toplevel makefile.
 *
 *	(C) Copyright 1992,1993 Richard Lyons
 *	      pclink@qus102.qld.tne.oz.au
 *	      rick@razorback.brisnet.org.au
 */

#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/termios.h>

#include <asm/dma.h>
#include <asm/system.h>
#include <asm/segment.h>

#define CONFIG_GS4500		/* should be in config.h */

#ifdef CONFIG_GS4500

#ifndef SCANNER_MAJOR
#define	SCANNER_MAJOR	30	/* major device */
#endif

#define	GS_DMA		3	/* DMA channel */
#define GS_DATA		0x272
#define GS_STATUS	0x273
#define GS_CONTROL	0x27A
#define GS_DMAHOLD	0x27B

static	int	isopen = 0;
static	int	resolution = -1;
static 	char	scanline[256], *p;
static	int	cnt = 0;
static	int	dpi[] = { 100, 200, 300, 400 };
static	int	bytesperline[] = { 52, 103, 153, 206 };
static	int	gs_mode[] = { 0xF0, 0xA0, 0x80, 0xE0 };

static struct scan_info {
	short	dpi;	/* dots per inch */
	short	bpl;	/* bytes per line */
	short	bpp;	/* bits per pixel */
} scan_info = { 0, 0, 0 };

static int scan1line(char *buf)
{
	cli();
	disable_dma(GS_DMA);
	clear_dma_ff(GS_DMA);
	set_dma_mode(GS_DMA, DMA_MODE_READ);
	set_dma_addr(GS_DMA, (unsigned int)buf);
	set_dma_count(GS_DMA, scan_info.bpl);
	enable_dma(GS_DMA);
	sti();
	outb(0, GS_DMAHOLD);
	while (!(inb(DMA1_STAT_REG) & 0x08))
		if (current->signal & ~current->blocked)
			return 0;
		else
			schedule();
	return scan_info.bpl;
}

static int gs4500_stat(void)
{
	int	tries, cc;

	outb(1, GS_CONTROL);
	outb(1, GS_DMAHOLD);
	for (tries = 100000; --tries && !(inb(GS_STATUS) & 0x80); ) ;
	if (tries)
		for (tries = 100000; --tries && ((cc = inb(GS_STATUS)) & 0x80); ) ;
	outb(0, GS_CONTROL);
	return tries?cc:-EIO;
}

static int gs4500_open(struct inode * inode, struct file * file)
{
	int	cc;

	if (isopen)
		return -EBUSY;
	if ((cc = gs4500_stat()) < 0)
		return cc;
	switch (cc & 0xA4) {			/* determine dpi */
		case 0x24:	resolution = 0;
				break;
		case 0x20:	resolution = 1;
				break;
		case 0x04:	resolution = 2;
				break;
		case 0x00:	resolution = 3;
				break;
		default:	return -EIO;
				break;
	}
	scan_info.dpi = dpi[resolution];
	scan_info.bpl = bytesperline[resolution];
	scan_info.bpp = 1;
	isopen = 1;
	cnt = 0;
	outb(gs_mode[resolution]|1, GS_CONTROL);/* turn scanner on */
	outb(1, GS_DMAHOLD);			/* disable DMA */
	return 0; 
}

static void gs4500_release(struct inode * inode, struct file * file)
{
	isopen = 0;
	outb(0, GS_CONTROL);
}

/*
 *	gs4500_read() and gs4500_ioctl() assume that they are only
 *	called after a successful gs4500_open().
 */

static int gs4500_read(struct inode * inode, struct file * file, char *buf, int size)
{
	int	scanned = 0, chars;

	while (size) {
		if (!cnt && (cnt = scan1line(p = scanline)) == 0)
			return scanned?scanned:-EIO;
		chars = cnt < size ? cnt : size;
		memcpy_tofs(buf, p, chars);
		size -= chars;
		cnt -= chars;
		p += chars;
		buf += chars;
		scanned += chars;
	}
	return scanned;
}

static int gs4500_ioctl(struct inode * inode, struct file * file, unsigned int cmd, unsigned long arg)
{
	if (cmd != TIOCGETD)
		return -EINVAL;
	if (arg)
		memcpy_tofs((char *)arg, (char *)&scan_info, sizeof(scan_info));
	return scan_info.bpl;
}

static struct file_operations gs4500_fops = {
        NULL,		/* seek */
	gs4500_read,	/* read */
	NULL,		/* write */
	NULL,		/* readdir */
	NULL,		/* select */
	gs4500_ioctl,	/* ioctl */
	NULL,		/* mmap */
        gs4500_open,	/* open */
        gs4500_release	/* release */
};

static struct file_operations *gs4500_init(void)
{
	if (gs4500_stat() >= 0) {
		if (request_dma(GS_DMA) == 0) {
			printk("GS4500 on DMA%d\n", GS_DMA);
			return &gs4500_fops;
		}
		printk("unable to get DMA%d for gs4500\n", GS_DMA);
	}
	return NULL;
}

static struct file_operations *(*scan_init[])(void) = {
	gs4500_init,
};

unsigned long scanner_init(unsigned long kmem_start)
{
	int	i;
	struct file_operations *f;

	for (i = 0; i < sizeof(scan_init) / sizeof(scan_init[0]); ++i)
		if ((f = (*scan_init[i])()) != NULL) {
			if (register_chrdev(SCANNER_MAJOR, "scanner", f))
				printk("unable to get major %d for scanner\n", SCANNER_MAJOR);
			break;
		}
	return kmem_start;
}
#else
unsigned long scanner_init(unsigned long kmem_start)
{
	return kmem_start;
}
#endif
